from xmlrpc.server import SimpleXMLRPCServer
from xmlrpc.client import ServerProxy, Fault  # 导入故障异常类Fault
from os.path import isfile, abspath, join  # 导入绝对路径的方法abspath
from urllib.parse import urlparse
import sys  # 导入系统模块

SimpleXMLRPCServer.allow_reuse_address = 1  # 保证节点服务器重启时能够立即访问
MAX_HISTORY_LENGTH = 6

UNHANDLED = 100  # 文件不存在的异常代码
ACCESS_DENIED = 200  # 文件访问受限的异常代码


class UnhandledQuery(Fault):  # 创建自定义异常类
    def __init__(self, message='无法处理请求！'):  # 定义构造方法
        Fault.__init__(self, UNHANDLED, message)  # 重载超类构造方法


class AccessDenied(Fault):  # 创建自定义异常类
    def __init__(self, message='访问资源受限！'):  # 定义构造方法
        Fault.__init__(self, ACCESS_DENIED, message)  # 重载超类构造方法


def inside(dir_path, file_path):  # 定义文件路径检查的方法
    directory = abspath(dir_path)  # 获取共享目录的绝对路径
    file = abspath(file_path)  # 获取请求资源的绝对路径
    return file.startswith(join(directory, ''))  # 返回请求资源的路径是否以共享目录路径开始


def get_port(url):
    result = urlparse(url)[1]
    port = result.split(':')[-1]
    return int(port)


class Node:
    def __init__(self, url, dir_name, secret):
        self.url = url
        self.dirname = dir_name
        self.secret = secret
        self.known = set()

    def _start(self):
        server = SimpleXMLRPCServer(('', get_port(self.url)), logRequests=False)
        server.register_instance(self)
        server.serve_forever()

    def _handle(self, filename):
        file_path = join(self.dirname, filename)
        if not isfile(file_path):  # 如果路径不是一个文件
            raise UnhandledQuery  # 抛出文件不存在的异常
        if not inside(self.dirname, file_path):  # 如果请求的资源不是共享目录中的资源
            raise AccessDenied  # 抛出访问资源受限异常
        return open(file_path).read()  # 未发生异常时返回读取的文件数据

    def _broadcast(self, filename, history):
        for other in self.known.copy():
            if other in history:
                continue
            try:
                server = ServerProxy(other)
                return server.query(filename, history)
            except Fault as f:  # 如果捕获访问故障异常获取异常代码
                if f.faultCode == UNHANDLED:  # 如果是文件不存在异常
                    pass  # 不做任何处理
                else:  # 如果是其它故障异常
                    self.known.remove(other)  # 从已知节点列表中移除节点
            except:  # 如果捕获其它异常（非故障异常）
                self.known.remove(other)  # 从已知节点列表中移除节点
        raise UnhandledQuery  # 如果已知节点都未能请求到资源，抛出文件不存在异常。

    def query(self, filename, history=[]):
        try:
            return self._handle(filename)
        except UnhandledQuery:  # 如果捕获文件不存在的异常
            history.append(self.url)
            if len(history) >= MAX_HISTORY_LENGTH:
                raise
            return self._broadcast(filename, history)

    def hello(self, other):
        self.known.add(other)
        return 0  # 必须返回非None的值

    def fetch(self, filename, secret):
        if secret != self.secret:  # 如果密钥不匹配
            raise AccessDenied  # 抛出访问资源受限异常
        result = self.query(filename)
        with open(join(self.dirname, filename), 'w') as file:
            file.write(result)
        return 0  # 必须返回非None的值


def main():  # 定义主程序函数，用于测试当前代码是否正常。
    url, directory, secret = sys.argv[1:]  # 获取通过命令行终端输入参数
    node = Node(url, directory, secret)  # 创建节点对象
    node._start()  # 启动节点服务器


if __name__ == '__main__':
    main()  # 运行主程序
